import {CashuMint, CashuWallet, getDecodedToken, getEncodedToken, type Proof} from "@cashu/cashu-ts";
import type {TokenEntry} from "@cashu/cashu-ts";
import {getMailSubscriptionDurationForSats} from "./general.ts";

class InvalidTokenException extends Error {
  constructor(message: string) {
    super(message);
    this.name = 'InvalidTokenException';
  }
}

export class TokenInfo {
  private static readonly ALLOWED_MINTS: readonly string[] = [
    'https://mint.minibits.cash/Bitcoin',
    'https://stablenut.umint.cash',
    'https://mint.lnvoltz.com',
    'https://mint.coinos.io',
    'https://mint.lnwallet.app',
    'https://mint.0xchat.com'
  ] as const;
  private static readonly MIN_AMOUNT = 21 as const;

  public readonly tokens: TokenEntry[];
  public readonly amount: number;
  public readonly mint: string;
  public readonly proofs: Proof[];

  constructor(protected readonly tokenString: string) {
    const decodedTokenData = getDecodedToken(tokenString);
    const tokens = decodedTokenData.token;
    const amount = tokens.flatMap(t => t.proofs).reduce((c, x) => c + x.amount, 0);

    if(decodedTokenData.token.length === 0)
      throw new InvalidTokenException('Invalid token format. We only accept tokens with at least one proof');
    if (tokens.slice(1).some(t => t.mint !== tokens[0].mint))
      throw new InvalidTokenException('Invalid token format. We only accept tokens with the same mint');
    if (!TokenInfo.ALLOWED_MINTS.includes(tokens[0].mint))
      throw new InvalidTokenException('Unsupported mint');
    if (amount < TokenInfo.MIN_AMOUNT)
      throw new InvalidTokenException(`Invalid amount. Minimum required: ${TokenInfo.MIN_AMOUNT} sats`);

    this.tokens = tokens;
    this.amount = amount;
    this.mint = tokens[0].mint;
    this.proofs = tokens.flatMap(t => t.proofs);
  }

  async receive(): Promise<string> {
    const mint = new CashuMint(this.mint);
    const wallet = new CashuWallet(mint);
    const newToken = await wallet.receive(this.tokenString);
    return getEncodedToken({
      token: [{
        mint: this.mint,
        proofs: newToken
      }]
    });
  }
}

export class TokenInfoWithMailSubscriptionDuration extends TokenInfo {
  public readonly duration: number;

  constructor(tokenString: string) {
    super(tokenString);
    this.duration = getMailSubscriptionDurationForSats(this.amount);
  }
}
